{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE MultiWayIf #-}
{-# LANGUAGE DerivingStrategies #-}

module Test.Core.LinearAlgebra where

import Data.Array.Accelerate hiding (Ord, Eq, map, (<=))
import Data.Array.Accelerate.Interpreter qualified as Naive
import Data.Bifunctor (first)
import Data.Bimap (Bimap)
import Data.Bimap qualified as Bimap
import Data.Map.Strict (Map)
import Data.Map.Strict qualified as M
import Gargantext.Core.LinearAlgebra qualified as LA
import Gargantext.Core.Methods.Matrix.Accelerate.Utils qualified as Legacy
import Gargantext.Core.Methods.Similarities.Accelerate.Distributional qualified as Legacy
import Gargantext.Core.Viz.Graph.Index qualified as Legacy
import Gargantext.Orphans.Accelerate (sliceArray)
import Prelude hiding ((^))
import Test.Tasty
import Test.Tasty.QuickCheck
import Data.Massiv.Array qualified as Massiv
import qualified Data.Array.Accelerate as A
import qualified Data.List.Split.Internals as Split
import Gargantext.Prelude (headMay)

--
-- Utility types and functions
--

newtype SquareMatrix a = SquareMatrix { _SquareMatrix :: Matrix a }
  deriving newtype (Show, Eq)

instance (Elt a, Show a, Arbitrary a) => Arbitrary (SquareMatrix a) where
  arbitrary = do
    x <- choose (0,30)
    let sh = Z :. x :. x
    SquareMatrix . fromList sh <$> vectorOf (x*x) arbitrary
  shrink = map (SquareMatrix) . sliceArray . _SquareMatrix

compareImplementations :: (Arbitrary a, Eq b, Show b)
                       => (a -> b)
                       -> (a -> c)
                       -> (c -> b)
                       -> a
                       -> Property
compareImplementations implementation1 implementation2 mapResults inputData
  = implementation1 inputData === mapResults (implementation2 inputData)

compareImplementations' :: (Arbitrary a, Eq c, Show c)
                       => (a -> b)
                       -> (a -> b)
                       -> (b -> c)
                       -> a
                       -> Property
compareImplementations' implementation1 implementation2 mapResults inputData
  = mapResults (implementation1 inputData) === mapResults (implementation2 inputData)

mapCreateIndices :: Ord t => (Map t Legacy.Index, Map Legacy.Index t) -> Bimap LA.Index t
mapCreateIndices (_m1, m2) = Bimap.fromList $ map (first LA.Index) $ M.toList m2

type TermDivNanShape = Z :. Int :. Int

twoByTwo :: Matrix Int
twoByTwo = fromList (Z :. 2 :. 2) (Prelude.replicate 4 0)

accelerate2MassivMatrix :: (Massiv.Unbox a, Elt a) => Matrix a -> Massiv.Matrix Massiv.D a
accelerate2MassivMatrix m =
  let (Z :. _r :. c) = A.arrayShape m
  in Massiv.delay $ Massiv.fromLists' @Massiv.U Massiv.Par $ Split.chunksOf c (A.toList m)

massiv2AccelerateMatrix :: Elt a => Massiv.Matrix Massiv.D a -> Matrix a
massiv2AccelerateMatrix m =
  let m' = Massiv.toLists2 m
      r  = Prelude.length m'
      c  = maybe 0 Prelude.length (headMay m')
  in A.fromList (Z :. r :. c) (mconcat m')

--
-- Main test runner
--

-- | Needed as the LLVM and Naive backend generates some double with a long exponent which
-- won't compare verbatim.
tests :: TestTree
tests = testGroup "LinearAlgebra" [
      testProperty "createIndices roundtrip" (compareImplementations (LA.createIndices @Int @Int) Legacy.createIndices mapCreateIndices)
    , testProperty "termDivNan"     compareTermDivNan
    , testGroup "distributional" [
        testProperty "2x2" (compareDistributional twoByTwo)
      , testProperty "roundtrips" (compareImplementations' @(SquareMatrix Int) (Legacy.distributionalWith Naive.run . _SquareMatrix)
                                                                               (Legacy.distributionalWith Naive.run . _SquareMatrix)
                                                                               id)
      ]
    ]

--
-- Tests
--

compareTermDivNan :: (Array TermDivNanShape Double)
                  -> (Array TermDivNanShape Double)
                  -> Property
compareTermDivNan i1 i2
  = let massiv     = LA.termDivNan (accelerate2MassivMatrix i1) (accelerate2MassivMatrix i2)
        accelerate = Naive.run (Legacy.termDivNan (use i1) (use i2))
    in accelerate === massiv2AccelerateMatrix massiv

compareDistributional :: Matrix Int
                      -> Property
compareDistributional i1
  = Legacy.distributionalWith Naive.run i1 === Legacy.distributionalWith Naive.run i1

